/*
 * $Id: cgiupl.c,v 1.4 1998/05/23 02:34:59 jbourne Exp $
 *
 * Based on code written by an anonymous software socialist.
 * 
 * Modified for use in CGI's by James Bourne <jbourne@hardrock.org>
 * May 9, 1998.  See the file COPYING for more information on Copyright
 *   
 *   Copyright (C) 1998  James Bourne <jbourne@hardrock.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>

#include <cgiupl.h>

#define CGIUPL_VERSION	"libcgiupl.a: version 0.2"

static void part_gets (char * buf, size_t size, int * error)
{
	unsigned char * out_ptr;
	size_t out_count;
	unsigned char ch;
	
	out_ptr = ((unsigned char *) buf)-1;
	out_count = 0;

	for (;;)
	{
		/* End of file processing */
		if (bytesRead >= contentLength)
		{
			*error = part_gets_FAIL;
			return;
		}
		
		/* Have we, or or we likely to, exceed the callers buffer size ? */
		if (out_count+1>= size)
		{
			*error = part_gets_FAIL;			
#ifdef _DEBUG
			printf ("<br>part_gets (truncated) line: %s\n", buf);
#endif
			return;
		}
		
		ch = (unsigned char) getc (stdin);
		bytesRead++;
		if (ch == '\r')
			continue;		 /* discard carriage-returns */
		out_ptr++;
		*out_ptr = ch;
		out_count++;
	
		/* Have we reached end of line */
		if (*out_ptr == '\n')
		{
			*error = part_gets_OK;
			if (out_ptr > (unsigned char *)buf && *(out_ptr - 1) == '\r')
				*(out_ptr - 1) = '\0';
			else	
				*out_ptr = '\0';
#ifdef _DEBUG
			printf ("<br>part_gets line: %s\n", buf);
#endif
			return;
		}
	}
}

static size_t part_read (void ** buf, size_t size, unsigned char * boundary,
	int * error)
{
	char * fileBuf;
	long out_ptr;
	size_t out_count;
	unsigned char ch;
	long candidate_EOP;
	int candidate_match_count;
	int boundary_len;
	
	boundary_len = strlen ((char *)boundary);
	if (size < boundary_len)
	{
		*error = part_read_FAIL;
		return 0;
	}
	
	fileBuf = (char *) *buf;	
	candidate_EOP = 0;
	candidate_match_count=0;
	out_ptr = -1;
	out_count = 0;

	for (;;)
	{
		/* End of file processing */
		if (bytesRead >= contentLength)
		{
			*error = part_read_FAIL;			
			*buf = fileBuf;
			return 0;
		}
		
		/* Have we, or or we likely to, exceed the callers buffer size ? */	
		if (out_count+boundary_len >= size)
		{
			if (candidate_EOP == 0)
			{
				/* we need room to store a whole new boundary */
				/* return the data we have so far */
				*error = part_read_OK;
				*buf = fileBuf;
				return out_count;
			}
			else
			{
				/* we're working on a candidate boundary, but it may end */
				/* with the start of a new candidate boundary */
				/* Allocate some room ... */
				size = size + boundary_len;
				fileBuf = (char*) realloc (fileBuf, size);
			}
		}

		out_ptr++;
		ch = (unsigned char) getc (stdin);
#ifdef _LOWDEBUG
		if (isprint((int)ch))
			printf (" $0x%x '%c'$ ", (int) ch, ch);
		else
			printf (" $0x%x '?'$ ", (int) ch);
#endif
		*(fileBuf+out_ptr) = ch;
		bytesRead++;
		out_count++;
	
		if (candidate_EOP == 0)
		{
			/* Does the current character start a boundary ? */
			if (ch == *boundary)
			{
				candidate_EOP = out_ptr;
				candidate_match_count = 1;
			}
			continue;
		}
	
		/* We have a candidate for a boundary match */
		candidate_match_count++;
		
		/* Does the candidate match the boundary (so far) ?  */	
#ifdef _LOWDEBUG
		printf ("<br>Boundary detection: bytesRead=%ld; match count=%d; "
			"stdin characters=%.*s; boundary characters=%.*s; ",
			bytesRead, candidate_match_count, 
			(int) candidate_match_count, fileBuf+candidate_EOP,
			(int) candidate_match_count, boundary);
#endif

		if (*(fileBuf+candidate_EOP+candidate_match_count-1) != 
			*(boundary +candidate_match_count-1))
		{
			if (ch == *boundary)
			{
				/* Could be the start of a new boundary */
				candidate_EOP = out_ptr;
				candidate_match_count = 1;
			}
			else	
			{
				candidate_EOP = 0;
			}
			continue;
		}
		
		/* If we haven't matched enough characters, loop ... */
		if (candidate_match_count < boundary_len)
			continue;
		
		if (candidate_match_count == boundary_len)
		{
			/* a full boundary match ? */
			char localBuf [128];
			int localError;
		
			/* Is this the end of stdin boundary? Anyway we'll bypass the cr lf */
			part_gets (localBuf, sizeof(localBuf), &localError);
			if (localError != part_gets_OK)
			{
				*error = part_read_FAIL;		   
				*buf = fileBuf;
				return 0;
			}
			if (strcmp(localBuf, "--")==0)
				*error = part_read_EOF;
			else
				*error = part_read_EOP;
		
			out_count=out_count - boundary_len; /* caller does not see boundary */
			*buf = fileBuf;
			return out_count;
		}
	}
}
	
/***************************************************************************/


int DecodeStdin (dentry *data)
{
	FILE *fUpload;
	char buf[1024];
	void * fileBuf;
	char boundaryCore[64];
	char boundaryFile[64];
	char boundaryGets[64];
	char boundaryLastGets[64];
	char fieldName[64];
	char fileName[1024];
	char tmpfileName[1024];
	char contentType[128];
	char *ptr, *ptr2;
	int error;                   
	int scanCount;
	enum {field, nonEmptyFile, emptyFile} part_type;
	int lastBoundaryFound;
	int num = 0;
	int len;
	
#ifdef _DEBUG
	printf("DecodeStdin: Entered<br>\n");
	fflush(stdout);
#endif

	if(contentLength <= 0)
		sscanf(getenv("CONTENT_LENGTH"), "%ld", &contentLength);

	ptr = getenv("CONTENT_TYPE");
	
	if(ptr != NULL)
	{
		ptr2 = strchr(ptr, ';');
		if(ptr2 != NULL)
			*ptr2 = '\0';
		
		if(strcmp(ptr, "multipart/form-data") != 0)
			return(ENCTYPE_FAIL);
	}
				
#ifdef _DEBUG
	printf("DecodeStdin: have content-type and content-length<br>\n");
	fflush(stdout);
#endif

/* Get boundary string declaration */
	part_gets (buf, sizeof(buf), &error);
	if (error != part_gets_OK)
		return(BOUNDGET_FAIL);

#ifdef _DEBUG
	printf("DecodeStdin: buf is %s<br>\n", buf);
	fflush(stdout);
#endif

	scanCount = sscanf (buf, "%*[-]%64[0123456789]", boundaryCore);
	if (scanCount != 1)
		return(BOUNDPRS_FAIL);
    
	strcpy (boundaryCore, buf);
	strcpy (boundaryFile, "\r\n");
	strcat (boundaryFile, boundaryCore);
	strcpy (boundaryGets, boundaryCore);
	sprintf(boundaryLastGets, "%s--", boundaryCore);

#ifdef _DEBUG
	printf("boundaryCore: %s<br>\n", boundaryCore);
	printf("boundaryFile: %s<br>\n", boundaryFile);
	printf("boundaryGets: %s<br>\n", boundaryGets);
	printf("boundaryLastGets: %s<p>\n", boundaryLastGets);
	fflush(stdout);
#endif

/* Process the multipart document */
	lastBoundaryFound=0;
	while (lastBoundaryFound == 0)
	{
		num++;
		/* Get the Content-Disposition line */
		part_gets (buf, sizeof(buf), &error);
		if (error == part_gets_EOF)
			break;

#ifdef _DEBUG
		printf("DecodeStdin: buf is %s<br>\n", buf);
		fflush(stdout);
#endif

		if (error != part_gets_OK)
			return(DISPGET_FAIL);
    
		/* Parse the Content-Disposition line */
		fieldName[0] = '\0';
		fileName[0] = '\0';
		scanCount = sscanf(buf,"Content-Disposition: form-data; "
			"name=\"%64[^\"]\"; "
			"filename=\"%1024[^\"]\""
			,fieldName, fileName);

		if (scanCount == 0)
			return(DISPPARSE_FAIL);

		/* What kind of part of the document have we found ? */    
		if (scanCount == 1)
		{
			if (strstr(buf, "filename=\"\"") != 0)
				part_type = emptyFile;
			else    
				part_type = field;
		}
		else
		{
			part_type = nonEmptyFile;
		}

		/* Get the line after the content disposition line */
		part_gets (buf, sizeof(buf), &error);
		if (error != part_gets_OK)
			return(DISPNEXT_FAIL);

		/* Process differently for each part type */
		switch (part_type)
		{
			case field:
			{
/* here we have to add a for loop to advance (for things like textarea) */
				/* Get the variable value */
				part_gets (buf, sizeof(buf), &error);
				if (error != part_gets_OK)
					return(FVALUE_FAIL);

				/* Extract the field value */
				data[num-1].name = malloc(strlen(fieldName)+1);
				strcpy(data[num-1].name, fieldName);
				data[num-1].val = malloc(strlen(buf)+3);
				sprintf(data[num-1].val, "%s\r\n", buf);
				data[num-1].type = TYPE_DATA;

				part_gets (buf, sizeof(buf), &error);
				if (error != part_gets_OK)
					return(VALPART_FAIL);
				len = strlen(buf)+1;
				while((strcmp(buf, boundaryCore) != 0) &&
					(strcmp(buf, boundaryLastGets) != 0))
				{
					len = len + strlen(buf) + 2;
					data[num-1].val = realloc(data[num-1].val, len);
					strncat(data[num-1].val, buf, len);
					strncat(data[num-1].val, "\r\n", len);

					/* Get the part boundary and check for EOF */
					part_gets (buf, sizeof(buf), &error);
					if (error != part_gets_OK)
						return(VALPART_FAIL);

				}
				/* trim the <CR><LF> from the end of the line */
				len = strlen(data[num-1].val);
				data[num-1].val[len-1] = '\0';
				data[num-1].val[len-2] = '\0';
				
				if (strcmp(buf, boundaryLastGets) == 0)
				{
					lastBoundaryFound = 1;
					continue;
				}
		
				break;
			}
			case emptyFile:
			{
				/* Get the empty line and ignore it */
				part_gets (buf, sizeof(buf), &error);
				if (error != part_gets_OK)
					return(NULLFILE_FAIL);
				
				/* Get the part boundary and check for EOF */
				part_gets (buf, sizeof(buf), &error);
				if (error != part_gets_OK)
					return(VALPART_FAIL);
				
				data[num-1].name = malloc(strlen(fieldName)+1);
				strcpy(data[num-1].name, fieldName);
				data[num-1].val = NULL;
				data[num-1].type = TYPE_EMPTYFILE;

				if (strcmp(buf, boundaryLastGets) == 0)
				{
			   	lastBoundaryFound = 1;
			   	continue;
			   }
		
				break;
			}
			
			case nonEmptyFile:
			{
				int i;
			
				/* Extract the Content-Type */ 
				contentType[0] = '\0';
				scanCount=sscanf(buf, "Content-Type: %128s", contentType); 

				/* Get the next line */
				part_gets (buf, sizeof(buf), &error);
				if (error != part_gets_OK)
					return(NNULLFILE_FAIL);
	
				/* Check for EOF */
				if (strcmp(buf, boundaryLastGets) == 0)
				{
			   	lastBoundaryFound = 1;
			   	continue;
			   }
		
				/* Make the upload file name OSS friendly  */
				for (i=0; fileName[i]; i++)
				{
					if((isalnum(fileName[i]) == 0) && (fileName[i] != '.')
						&& (fileName[i] != '-') && (fileName[i] != '_'))
						fileName[i]='_';
				}

				/* Add MIME type to protect against user uploading CGI code */	 
				if (strlen(fileName) > PATH_MAX)
					return(ENAMETOOLONG);
				
				if(cgiUploadDir[0] == '\0')
					strcpy(cgiUploadDir, "/tmp/");
				sprintf(tmpfileName, "%s/%s", cgiUploadDir, fileName);
				/* Open the output file */

				data[num-1].name = malloc(strlen(fieldName)+1);
				strcpy(data[num-1].name, fieldName);
				data[num-1].val = malloc(strlen(fileName)+1);
				strcpy(data[num-1].val, fileName);
				data[num-1].type = TYPE_FILE;

				if((fUpload  = fopen(tmpfileName, "w" )) == NULL )
					return(errno);

				chmod(tmpfileName, S_IRUSR|S_IWUSR);

				/* Copy the part to the file */
				for (;;)
				{
					size_t bytesReturned;
				
					fileBuf = malloc (1024);
					bytesReturned=part_read 
							  (&fileBuf, 1024, 
							  (unsigned char *) boundaryFile, &error);
					if (error == part_read_OK || 
						error == part_read_EOP||
						error == part_read_EOF) 
						fwrite (fileBuf,1,bytesReturned,fUpload); 
					free (fileBuf);	
					if (error == part_read_OK)
						continue;   /* there's more data */ 
				   
					/* We're finished with the file */
					fclose (fUpload);
					fUpload = NULL;
					/* Check the error code to see what to do next */
					if (error == part_read_FAIL)
						return(error);

					if (error == part_read_EOF) 
						lastBoundaryFound = 1;
				
					break;
				}
					
				break;
			}
		}
	}
	return(num);
}

char *check_dserror(int error)
{
	switch(error)
	{
		case 0:
			return("No Error");
			break;
		case part_read_EOF:
			return("Partition read error, EOF occured");
			break;
		case part_read_EOP:
			return("Partition read error, EOP occured");
			break;
		case part_read_FAIL:
			return("Partition read failed");
			break;
		case part_gets_EOF:
			return("Partition gets error, EOF occured");
			break;
		case part_gets_FAIL:
			return("Partition gets failed");
			break;
		case BOUNDGET_FAIL:
			return("Error getting boundry declaration");
			break;
		case BOUNDPRS_FAIL:
			return("Error parsing boundry declaration");
			break;
		case DISPPARSE_FAIL:
			return("Error parsing Content-Disposition");
			break;
		case DISPGET_FAIL:
			return("Error getting the dispostition line");
			break;
		case DISPNEXT_FAIL:
			return("Error getting line after disposition line");
			break;
		case FVALUE_FAIL:
			return("Error getting field value");
			break;
		case VALPART_FAIL:
			return("Error getting partition after value");
			break;
		case NULLFILE_FAIL:
			return("Error getting null line for empty file");
			break;
		case NNULLFILE_FAIL:
			return("Error getting null line for nonempty file");
			break;
		case ENCTYPE_FAIL:
			return("Form ENCTYPE is not multipart/form-data");
			break;
		default:
			return("Unknown Error");
	}
}
